# pylint: disable=protected-access, invalid-name, too-many-locals
import secrets
import base64
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.bindings.openssl.binding import Binding
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from django.utils.crypto import salted_hmac

copenssl = Binding.lib
cffi = Binding.ffi


# SMIME isn't supported by pyca/cryptography:
# https://github.com/pyca/cryptography/issues/1621
def pkcs7_sign(certcontent,
               keycontent,
               wwdr_certificate,
               data,
               key_password=None,
               flag=copenssl.PKCS7_BINARY | copenssl.PKCS7_DETACHED):
    """Sign data with PKCS#7.

    Args:
        keycontent (bytes): Content of pem file certificate
        keycontent (bytes): Content of key file
        wwdr_certificate (bytes): Content of Intermediate cert file
        data (bytes): Data to be signed
        key_password (bytes, optional): key file passwd. Defaults to None.
        flag (int, optional): Flags to be passed to PKCS7_sign C lib.
            Defaults to copenssl.PKCS7_BINARY|copenssl.PKCS7_DETACHED.
    """

    backend = default_backend()

    # Load cert and key
    pkey = load_pem_private_key(keycontent, key_password, backend=backend)
    cert = x509.load_pem_x509_certificate(certcontent, backend=backend)

    # Load intermediate cert and push it into < Cryptography_STACK_OF_X509 * >
    intermediate_cert = x509.load_der_x509_certificate(
        wwdr_certificate,
        backend,
    )
    certs_stack = copenssl.sk_X509_new_null()
    # https://www.openssl.org/docs/man1.1.1/man3/sk_TYPE_push.html
    # int sk_TYPE_push(STACK_OF(TYPE) *sk, const TYPE *ptr);
    # return amount of certs into certs_stack, -1 on error
    # TODO: raise on count < 1
    _count = copenssl.sk_X509_push(certs_stack, intermediate_cert._x509)

    bio = backend._bytes_to_bio(data)
    # From
    # pyca/cryptography/src/_cffi_src/openssl/pkcs7.py
    # PKCS7 *PKCS7_sign(X509 *, EVP_PKEY *,
    #                   Cryptography_STACK_OF_X509 *, BIO *, int);
    # signing-time attr is automatically added:
    # https://www.openssl.org/docs/man1.1.1/man3/PKCS7_sign.html
    pkcs7 = copenssl.PKCS7_sign(
        cert._x509,
        pkey._evp_pkey,
        certs_stack,
        bio.bio,
        flag,
    )

    bio_out = backend._create_mem_bio_gc()
    copenssl.i2d_PKCS7_bio(bio_out, pkcs7)

    signed_pkcs7 = backend._read_mem_bio(bio_out)
    return signed_pkcs7


def gen_random_token():
    rand1 = secrets.token_bytes(16)
    rand2 = secrets.token_bytes(7)
    rand2 = salted_hmac(rand1, rand2).digest()
    part_a = base64.urlsafe_b64encode(rand2).rstrip(b'=').decode('ascii')
    part_b = secrets.token_urlsafe(20)
    return part_b + part_a



from django.utils.module_loading import import_string
from django_walletpass.settings import dwpconfig as WALLETPASS_CONF

STORAGE_CLASS = import_string(WALLETPASS_CONF['STORAGE_CLASS'])


class WalletPassStorage(STORAGE_CLASS):
    pass



import os
import uuid
import hashlib
import json
import tempfile
import secrets
import zipfile
from glob import glob
from django.core.exceptions import ValidationError
from django.utils.module_loading import import_string
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django_walletpass import crypto
from django_walletpass.storage import WalletPassStorage
from django_walletpass.files import WalletpassContentFile
from django_walletpass.settings import dwpconfig as WALLETPASS_CONF


class PassBuilder:
    pass_data = {}
    pass_data_required = {
        "passTypeIdentifier": WALLETPASS_CONF['PASS_TYPE_ID'],
        "serialNumber": None,
        "teamIdentifier": WALLETPASS_CONF['TEAM_ID'],
        "webServiceURL": WALLETPASS_CONF['SERVICE_URL'],
        "authenticationToken": None,
    }
    directory = None
    extra_files = {}
    manifest_dict = {}
    builded_pass_content = None

    def __init__(self, directory=None):
        self.directory = directory
        if directory is not None:
            self._load_pass_json_file_if_exists(directory)
        self.pass_data_required.update({
            "serialNumber": secrets.token_urlsafe(20),
            "authenticationToken": crypto.gen_random_token(),
        })

    def _copy_dir_files(self, tmp_pass_dir):
        """Copy files from provided base dir to temporal dir

        Args:
            tmp_pass_dir (str): temporal dir path
        """
        for absolute_filepath in glob(os.path.join(self.directory, '**'), recursive=True):
            filename = os.path.basename(absolute_filepath)
            relative_file_path = os.path.relpath(absolute_filepath, self.directory)
            if filename == '.DS_Store':
                continue
            if not os.path.isfile(absolute_filepath):
                continue
            filecontent = open(absolute_filepath, 'rb').read()
            # Add files to manifest
            self.manifest_dict[relative_file_path] = hashlib.sha1(filecontent).hexdigest()
            dest_abs_filepath = os.path.join(tmp_pass_dir, relative_file_path)
            dest_abs_dirpath = os.path.dirname(dest_abs_filepath)
            if not os.path.exists(dest_abs_dirpath):
                os.makedirs(dest_abs_dirpath)
            ff = open(dest_abs_filepath, 'wb')
            ff.write(filecontent)
            ff.close()

    def _write_extra_files(self, tmp_pass_dir):
        """Write extra files contained in self.extra_files into tmp dir

        Args:
            tmp_pass_dir (str): temporal dir path
        """
        for relative_file_path, filecontent in self.extra_files.items():
            # Add files to manifest
            self.manifest_dict[relative_file_path] = hashlib.sha1(filecontent).hexdigest()
            dest_abs_filepath = os.path.join(tmp_pass_dir, relative_file_path)
            dest_abs_dirpath = os.path.dirname(dest_abs_filepath)
            if not os.path.exists(dest_abs_dirpath):
                os.makedirs(dest_abs_dirpath)
            ff = open(dest_abs_filepath, 'wb')
            ff.write(filecontent)
            ff.close()

    def _write_pass_json(self, tmp_pass_dir):
        """Write content of self.pass_data to pass.json (in JSON format)

        Args:
            tmp_pass_dir (str): temporal dir path where pass.json will be saved
        """
        pass_json = json.dumps(self.pass_data)
        pass_json_bytes = bytes(pass_json, 'utf8')
        # Add pass.json to manifest
        self.manifest_dict['pass.json'] = hashlib.sha1(pass_json_bytes).hexdigest()
        ff = open(os.path.join(tmp_pass_dir, 'pass.json'), 'wb')
        ff.write(pass_json_bytes)
        ff.close()

    def _write_manifest_json_and_signature(self, tmp_pass_dir):
        """Write the content of self.manifest_dict into manifest.json

        Args:
            tmp_pass_dir (str): temporal dir path
        """
        manifest_json = json.dumps(self.manifest_dict)
        manifest_json_bytes = bytes(manifest_json, 'utf8')
        ff = open(os.path.join(tmp_pass_dir, 'manifest.json'), 'wb')
        ff.write(manifest_json_bytes)
        ff.close()
        signature_content = crypto.pkcs7_sign(
            certcontent=WALLETPASS_CONF['CERT_CONTENT'],
            keycontent=WALLETPASS_CONF['KEY_CONTENT'],
            wwdr_certificate=WALLETPASS_CONF['WWDRCA_CONTENT'],
            data=manifest_json_bytes,
            key_password=WALLETPASS_CONF['KEY_PASSWORD'],
        )
        ff = open(os.path.join(tmp_pass_dir, 'signature'), 'wb')
        ff.write(signature_content)
        ff.close()

    def _zip_all(self, directory):
        zip_file_path = os.path.join(directory, '..', 'walletcard.pkpass')
        zip_pkpass = zipfile.ZipFile(zip_file_path, 'w', zipfile.ZIP_DEFLATED)
        for filepath in glob(os.path.join(directory, '**'), recursive=True):
            relative_file_path = os.path.relpath(filepath, directory)
            zip_pkpass.write(filepath, arcname=relative_file_path)
        zip_pkpass.close()
        return open(zip_file_path, 'rb').read()

    def _load_pass_json_file_if_exists(self, directory):
        """Call self.load_pass_json_file if pass.json exist

        Args:
            directory (str): directory where pass.json resides
        """
        if os.path.isfile(os.path.join(directory, 'pass.json')):
            self.load_pass_json_file(directory)

    def _clean_manifest(self):
        self.manifest_dict = {}

    def _clean_builded_pass_content(self):
        self.builded_pass_content = None

    def validate(self):
        """Some validations before build the .pkpass file

        Raises:
            ValidationError: on validation error
        """
        if not self.pass_data:
            raise ValidationError(_("Cannot obtain data for pass.json."))

    def clean(self):
        self._clean_manifest()
        self._clean_builded_pass_content()
        self.validate()

    def load_pass_json_file(self, dir):
        """Load json file without test if exists.

        Args:
            dir (str): path where resides the pass.json
        """
        json_data = open(os.path.join(dir, 'pass.json'), 'r').read()
        self.pass_data = json.loads(json_data)

    def pre_build_pass_data(self):
        """Update self.pass_data with self.pass_data_required content
        """
        self.pass_data.update(self.pass_data_required)

    def build(self):
        """Build .pkpass file
        """
        self.clean()
        with tempfile.TemporaryDirectory() as tmpdirname:
            os.mkdir(os.path.join(tmpdirname, 'data.pass'))
            tmp_pass_dir = os.path.join(tmpdirname, 'data.pass')
            if self.directory:
                self._copy_dir_files(tmp_pass_dir)
            self._write_extra_files(tmp_pass_dir)
            self.pre_build_pass_data()
            self._write_pass_json(tmp_pass_dir)
            self._write_manifest_json_and_signature(tmp_pass_dir)
            self.builded_pass_content = self._zip_all(tmp_pass_dir)
        return self.builded_pass_content

    def write_to_model(self, instance=None):
        """Saves the content of builded and zipped pass into Pass model.

        Args:
            instance (Pass, optional): Pass instance, a new one will be created
                if none provided. Defaults to None.

        Returns:
            Pass: instance of Pass (already saved)
        """
        if instance is None:
            instance = Pass()

        setattr(instance, 'pass_type_identifier', WALLETPASS_CONF['PASS_TYPE_ID'])
        setattr(instance, 'serial_number', self.pass_data_required.get('serialNumber'))
        setattr(instance, 'authentication_token', self.pass_data_required.get('authenticationToken'))

        if instance.data.name:
            filename = os.path.basename(instance.data.name)
        else:
            filename = f"{uuid.uuid1()}.pkpass"

        content = WalletpassContentFile(self.builded_pass_content)
        instance.data.delete()
        instance.data.save(filename, content)

        return instance

    def add_file(self, path, content):
        self.extra_files[path] = content


class Pass(models.Model):
    """
    Pass instance
    """
    pass_type_identifier = models.CharField(max_length=150)
    serial_number = models.CharField(max_length=150)
    authentication_token = models.CharField(max_length=150)
    data = models.FileField(
        upload_to=WALLETPASS_CONF['UPLOAD_TO'],
        storage=WalletPassStorage(),
    )
    updated_at = models.DateTimeField(auto_now=True)

    def push_notification(self):
        klass = import_string(WALLETPASS_CONF['WALLETPASS_PUSH_CLASS'])
        push_module = klass()
        for registration in self.registrations.all():
            push_module.push_notification_from_instance(registration)

    def new_pass_builder(self, directory=None):
        builder = PassBuilder(directory)
        builder.pass_data_required.update({
            "passTypeIdentifier": self.pass_type_identifier,
            "serialNumber": self.serial_number,
            "authenticationToken": self.authentication_token,
        })
        return builder

    def get_pass_builder(self):
        builder = PassBuilder()
        with tempfile.TemporaryDirectory() as tmpdirname:
            os.mkdir(os.path.join(tmpdirname, 'data.pass'))
            tmp_pass_dir = os.path.join(tmpdirname, 'data.pass')
            # Put zip file into tmp dir
            zip_path = os.path.join(tmpdirname, 'walletcard.pkpass')
            zip_pkpass = open(zip_path, 'wb')
            zip_pkpass.write(self.data.read())
            zip_pkpass.close()
            # Extract zip file to tmp dir
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                zip_ref.extractall(tmp_pass_dir)
            # Populate builder with zip content
            for filepath in glob(os.path.join(tmp_pass_dir, '**'), recursive=True):
                filename = os.path.basename(filepath)
                relative_file_path = os.path.relpath(filepath, tmp_pass_dir)
                if filename == 'pass.json':
                    builder.load_pass_json_file(tmp_pass_dir)
                    continue
                if relative_file_path in ['signature', 'manifest.json', '.', '..']:
                    continue
                if not os.path.isfile(filepath):
                    continue
                builder.add_file(relative_file_path, open(filepath, 'rb').read())
        # Load of these fields due to that those fields are ignored
        # on pass.json loading
        builder.pass_data_required.update({
            "passTypeIdentifier": self.pass_type_identifier,
            "serialNumber": self.serial_number,
            "authenticationToken": self.authentication_token,
        })
        return builder

    def __unicode__(self):
        return self.serial_number

    class Meta:
        verbose_name_plural = "passes"
        unique_together = (
            'pass_type_identifier',
            'serial_number',
        ),


class Registration(models.Model):
    """
    Registration of a Pass on a device
    """
    device_library_identifier = models.CharField(max_length=150)
    push_token = models.CharField(max_length=150)
    pazz = models.ForeignKey(
        Pass,
        on_delete=models.CASCADE,
        related_name='registrations',
    )

    def __unicode__(self):
        return self.device_library_identifier


class Log(models.Model):
    """
    Log message sent by a device
    """
    message = models.TextField()

    def __unicode__(self):
        return self.message



# Generated by Django 2.2 on 2019-06-11 13:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_walletpass', '0003_auto_20190610_1434'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pass',
            name='authentication_token',
            field=models.CharField(max_length=150),
        ),
    ]



# Generated by Django 2.2 on 2019-06-13 16:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_walletpass', '0005_auto_20190613_1433'),
    ]

    operations = [
        migrations.AlterField(
            model_name='registration',
            name='pazz',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='registrations', to='django_walletpass.Pass'),
        ),
    ]



from django.test import TestCase
from django_walletpass import crypto
from django_walletpass.models import PassBuilder
from django_walletpass.settings import dwpconfig as WALLETPASS_CONF


class CryptoTestCase(TestCase):
    def test_smime_sign(self):
        crypto.pkcs7_sign(
            certcontent=WALLETPASS_CONF['CERT_CONTENT'],
            keycontent=WALLETPASS_CONF['KEY_CONTENT'],
            wwdr_certificate=WALLETPASS_CONF['WWDRCA_CONTENT'],
            data=b'data to be signed',
            key_password=WALLETPASS_CONF['KEY_PASSWORD'],
        )


class BuilderTestCase(TestCase):
    def test_build_pkpass(self):
        builder = PassBuilder()
        builder.pass_data = {
            "formatVersion": 1,
            "barcode": {
                "message": "123456789",
                "format": "PKBarcodeFormatPDF417",
                "messageEncoding": "iso-8859-1"
            },
            "organizationName": "Organic Produce",
            "description": "Organic Produce Loyalty Card",
            "logoText": "Organic Produce",
            "foregroundColor": "rgb(255, 255, 255)",
            "backgroundColor": "rgb(55, 117, 50)",
        }

        builder.build()

        instance = builder.write_to_model()
        instance.save()
        self.assertIsNotNone(instance.pk)

        builder2 = instance.get_pass_builder()
        builder2.build()
        self.assertEqual(builder.manifest_dict, builder2.manifest_dict)
        self.assertEqual(builder.pass_data, builder2.pass_data)

        builder2.pass_data.update({"organizationName": 'test'})
        builder2.build()
        builder2.write_to_model(instance)
        instance.save()

        builder3 = instance.get_pass_builder()
        builder3.build()

        self.assertEqual(builder2.manifest_dict, builder3.manifest_dict)
        self.assertEqual(builder2.pass_data, builder3.pass_data)

        self.assertNotEqual(builder.manifest_dict, builder3.manifest_dict)
        self.assertNotEqual(builder.pass_data, builder3.pass_data)



from django.db.models.signals import post_save
from django.utils.module_loading import import_string
from django.dispatch import receiver
from django_walletpass.models import Pass
from django_walletpass.settings import dwpconfig as WALLETPASS_CONF


@receiver(post_save, sender=Pass)
def fill_total(instance=None, **_kwargs):
    klass = import_string(WALLETPASS_CONF['WALLETPASS_PUSH_CLASS'])
    push_module = klass()
    for registration in instance.registrations.all():
        push_module.push_notification_from_instance(registration)



import os
from collections import UserDict
from django.conf import settings as django_settings
from django.test.signals import setting_changed

FULL_BASE_DIR = os.path.dirname(os.path.abspath(__file__))
DEFAULTS = {
    'CERT_PATH': None,
    'CERT_CONTENT': None,
    'KEY_CONTENT': None,
    'WWDRCA_CONTENT': None,
    'KEY_PATH': None,
    'KEY_PASSWORD': None,
    'APPLE_WWDRCA_CERT_PATH': os.path.join(FULL_BASE_DIR, 'certs', 'AppleWWDRCA.cer'),
    'WWDRCA_CONTENT': open(os.path.join(FULL_BASE_DIR, 'certs', 'AppleWWDRCA.cer'), 'rb').read(),
    'APPLE_WWDRCA_PEM_PATH': os.path.join(FULL_BASE_DIR, 'certs', 'AppleWWDRCA.pem'),
    'PASS_TYPE_ID': None,
    'TEAM_ID': None,
    'SERVICE_URL': None,
    'WALLETPASS_PUSH_CLASS': 'django_walletpass.services.PushBackend',
    'PUSH_SANDBOX': False,
    'STORAGE_CLASS': django_settings.DEFAULT_FILE_STORAGE,
    'UPLOAD_TO': 'passes'
}


class ConfigManager(UserDict):
    def update_conf(self):
        self.data = DEFAULTS
        new = django_settings.WALLETPASS
        if 'CERT_PATH' in new:
            new['CERT_CONTENT'] = open(new['CERT_PATH'], 'rb').read()
        if 'KEY_PATH' in new:
            new['KEY_CONTENT'] = open(new['KEY_PATH'], 'rb').read()
        if 'APPLE_WWDRCA_CERT_PATH' in new:
            new['WWDRCA_CONTENT'] = open(new['APPLE_WWDRCA_CERT_PATH'], 'rb').read()
        self.data.update(new)


dwpconfig = ConfigManager(DEFAULTS)
dwpconfig.update_conf()


def update_conf(*args, **kwargs):
    if kwargs['setting'] == 'WALLETPASS':
        dwpconfig.update_conf()


setting_changed.connect(update_conf)



# Generated by Django 2.2 on 2019-06-13 14:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_walletpass', '0004_auto_20190611_1306'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pass',
            name='pass_type_identifier',
            field=models.CharField(max_length=150),
        ),
        migrations.AlterField(
            model_name='pass',
            name='serial_number',
            field=models.CharField(max_length=150),
        ),
        migrations.AlterField(
            model_name='registration',
            name='device_library_identifier',
            field=models.CharField(max_length=150),
        ),
        migrations.AlterField(
            model_name='registration',
            name='push_token',
            field=models.CharField(max_length=150),
        ),
    ]



# Generated by Django 2.2 on 2019-06-10 14:34

from django.db import migrations, models
import django_walletpass.storage


class Migration(migrations.Migration):

    dependencies = [
        ('django_walletpass', '0002_auto_20190429_1819'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pass',
            name='data',
            field=models.FileField(storage=django_walletpass.storage.WalletPassStorage(), upload_to='passes'),
        ),
    ]



from django.conf.urls import url
from . import classviews

urlpatterns = [
    url(
        r'^v1/devices/(?P<device_library_id>.+)/registrations/(?P<pass_type_id>.+)/(?P<serial_number>.+)$',
        classviews.RegisterPassViewSet.as_view({
            'post': 'create',
            'delete': 'destroy',
        }),
        name='walletpass_register_pass',
    ),
    url(
        r'^v1/devices/(?P<device_library_id>.+)/registrations/(?P<pass_type_id>.+)$',
        classviews.RegistrationsViewSet.as_view({
            'get': 'list',
        }),
        name='walletpass_registrations',
    ),
    url(
        r'^v1/passes/(?P<pass_type_id>.+)/(?P<serial_number>.+)$',
        classviews.LatestVersionViewSet.as_view({
            'get': 'retrieve'
        }),
        name='walletpass_latest_version',
    ),
    url(
        r'^v1/log$',
        classviews.LogViewSet.as_view({
            'post': 'create',
        }),
        name='walletpass_log',
    ),
]



# Generated by Django 2.2 on 2019-04-29 18:19

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_walletpass', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='pass',
            options={'verbose_name_plural': 'passes'},
        ),
    ]



import json
from datetime import datetime
from calendar import timegm
from django.http import HttpResponse
from django.utils.http import http_date
from django.middleware.http import ConditionalGetMiddleware
from django.db.models import Max
import django.dispatch
from django.shortcuts import get_object_or_404
from rest_framework import viewsets, status
from rest_framework.response import Response
from rest_framework.permissions import AllowAny
from django_walletpass.models import Pass, Registration, Log

FORMAT = '%Y-%m-%d %H:%M:%S'
PASS_REGISTERED = django.dispatch.Signal()
PASS_UNREGISTERED = django.dispatch.Signal()


def get_pass(pass_type_id, serial_number):
    return get_object_or_404(Pass, pass_type_identifier=pass_type_id, serial_number=serial_number)


class RegistrationsViewSet(viewsets.ViewSet):
    """
    Gets the Serial Numbers for Passes Associated with a Device
    """
    permission_classes = (AllowAny, )

    def list(self, request, device_library_id, pass_type_id):
        passes = Pass.objects.filter(
            registrations__device_library_identifier=device_library_id,
            pass_type_identifier=pass_type_id,
        )
        if passes.count() == 0:
            return Response({}, status=status.HTTP_400_BAD_REQUEST)

        if 'passesUpdatedSince' in request.GET:
            passes = passes.filter(updated_at__gt=datetime.strptime(request.GET['passesUpdatedSince'], FORMAT))

        if passes:
            last_updated = passes.aggregate(Max('updated_at'))['updated_at__max']
            serial_numbers = [p.serial_number for p in passes.filter(updated_at=last_updated).all()]
            response_data = {'lastUpdated': last_updated.strftime(FORMAT), 'serialNumbers': serial_numbers}
            return Response(response_data)

        return Response({}, status=status.HTTP_204_NO_CONTENT)


# TODO: Refactor.
# - Use a ModelViewSet
# - Use auth class
class RegisterPassViewSet(viewsets.ViewSet):
    """
    get: Registers a Device to Receive Push Notifications for a Pass
    destroy: Unregister a Device
    """
    permission_classes = (AllowAny, )

    def create(self, request, device_library_id, pass_type_id, serial_number):

        pass_ = get_pass(pass_type_id, serial_number)
        if request.META.get('HTTP_AUTHORIZATION') != 'ApplePass %s' % pass_.authentication_token:
            return Response({}, status=status.HTTP_401_UNAUTHORIZED)
        registration = Registration.objects.filter(
            device_library_identifier=device_library_id,
            pazz=pass_,
        )
        if registration:
            return Response({}, status=status.HTTP_200_OK)
        body = json.loads(request.body)
        new_registration = Registration(
            device_library_identifier=device_library_id,
            push_token=body['pushToken'],
            pazz=pass_,
        )
        new_registration.save()
        PASS_REGISTERED.send(sender=pass_)
        return Response({}, status=status.HTTP_201_CREATED)

    def destroy(self, request, device_library_id, pass_type_id, serial_number):
        pass_ = get_pass(pass_type_id, serial_number)
        if request.META.get('HTTP_AUTHORIZATION') != 'ApplePass %s' % pass_.authentication_token:
            return Response({}, status=status.HTTP_401_UNAUTHORIZED)
        registration = Registration.objects.filter(
            device_library_identifier=device_library_id,
            pazz=pass_,
        )
        if registration:
            return Response({}, status=status.HTTP_200_OK)
        registration.delete()
        PASS_UNREGISTERED.send(sender=pass_)
        return Response({}, status=status.HTTP_200_OK)


class LatestVersionViewSet(viewsets.ViewSet):
    """
    get: Gets the latest version of a Pass
    """
    permission_classes = (AllowAny, )

    def retrieve(self, request, pass_type_id, serial_number):
        pass_ = get_pass(pass_type_id, serial_number)

        if request.META.get('HTTP_AUTHORIZATION') != 'ApplePass %s' % pass_.authentication_token:
            return Response({}, status=status.HTTP_401_UNAUTHORIZED)

        # TODO: handle remote storages like S3
        response = HttpResponse(pass_.data.read(), content_type='application/vnd.apple.pkpass')
        response['Content-Disposition'] = 'attachment; filename=pass.pkpass'

        response['Last-Modified'] = http_date(timegm(pass_.updated_at.utctimetuple()))
        middleware = ConditionalGetMiddleware()
        return middleware.process_response(request, response)


# TODO: use ModelViewSet
class LogViewSet(viewsets.ViewSet):
    """
    Logs messages from devices
    """
    permission_classes = (AllowAny, )

    def create(self, request):
        json_body = json.loads(request.body)
        for message in json_body['logs']:
            Log(message=message).save()
        return Response({}, status=status.HTTP_200_OK)



import logging
from ssl import SSLError
from hyper.tls import init_context
from hyper.http20.exceptions import StreamResetError
from apns2.client import APNsClient
from apns2.credentials import Credentials
from apns2.payload import Payload
from django_walletpass.models import Registration
from django_walletpass.settings import dwpconfig as WALLETPASS_CONF

logger = logging.getLogger('django.contrib.gis')


class PushBackend:
    def push_notification(self, credentials, push_sandbox, pass_type_id, push_token):
        payload = Payload()
        try:
            client = APNsClient(
                credentials,
                use_sandbox=push_sandbox,
                use_alternative_port=False,
            )
            client.send_notification(
                push_token,
                payload,
                pass_type_id,
            )
        except SSLError as e:
            logger.error("django_walletpass SSLError: %s", e)
        except StreamResetError as e:
            logger.error("django_walletpass StreamResetError. Bad cert or token? %s", e)

    def get_credentials(self):
        context = init_context(
            cert=(
                WALLETPASS_CONF['CERT_PATH'],
                WALLETPASS_CONF['KEY_PATH'],
            ),
            # cert_path=WALLETPASS_CONF['APPLE_WWDRCA_PEM_PATH'],
            cert_password=WALLETPASS_CONF['KEY_PASSWORD'],
        )

        return Credentials(context)

    def push_notification_with_token(self, token):
        self.push_notification(
            self.get_credentials(),
            push_sandbox=WALLETPASS_CONF['PUSH_SANDBOX'],
            pass_type_id=WALLETPASS_CONF['PASS_TYPE_ID'],
            push_token=token,
        )

    def push_notification_from_instance(self, registration_instance):
        self.push_notification_with_token(registration_instance.push_token)

    def push_notification_from_pk(self, registration_pk):
        registration = Registration.objects.get(pk=registration_pk)
        return self.push_notification_from_instance(registration)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()

setup(
    name='django-walletpass',
    python_requires='>=3.5.0',
    version='0.11',
    author='Develatio Technologies S.L.',
    author_email='contacto@develat.io',
    packages=find_packages(),
    include_package_data=True,
    url='http://github.com/develatio/django-walletpass/',
    license='BSD',
    install_requires=[
        'Django>=2.0',
        'cryptography>=2.4.2',
        'apns2>=0.5.0',
        'djangorestframework>=3.8',
    ],
    description='Django .pkpass builder, server and push notifications',
    long_description=README,
    long_description_content_type='text/markdown',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Internationalization',
        'Topic :: Software Development :: Localization',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
)



# Generated by Django 2.2 on 2019-06-13 18:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_walletpass', '0006_auto_20190613_1607'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pass',
            name='updated_at',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# Generated by Django 2.2 on 2019-04-29 18:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Log',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Pass',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('pass_type_identifier', models.CharField(max_length=50)),
                ('serial_number', models.CharField(max_length=50)),
                ('authentication_token', models.CharField(max_length=50)),
                ('data', models.FileField(upload_to='passes')),
                ('updated_at', models.DateTimeField()),
            ],
            options={
                'unique_together': {('pass_type_identifier', 'serial_number')},
            },
        ),
        migrations.CreateModel(
            name='Registration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('device_library_identifier', models.CharField(max_length=50)),
                ('push_token', models.CharField(max_length=50)),
                ('pazz', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_walletpass.Pass')),
            ],
        ),
    ]
